#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>

#include "explore_param.h"

#define NUM_REFLECTIONS 1

#define MPIROOT         0

#define MISFIDX(x1,x2,a) ((x1)*(nparam*2) + \
                          (x2)*(2) + a)

double vectors_dot_prod(const double *x, const double *y, int n);

void matrix_vector_mult(const double *mat, const double *vec, double *result, int rows, int cols, double alpha);

int main( int argc, char* argv[] )
{
  int  ppw;  // Points per wavelength
  int  N;   
  int  nx;   // nx velocity points 
  int  nx1;  // nx + 1 stress points
  int  reflections_times[NUM_REFLECTIONS] = {6}; // must be even: 2,4,6,8,...; [2 4 6 8 10 15000] 
  int  lambda_propagated[NUM_REFLECTIONS];
  int  nt[NUM_REFLECTIONS];

  float   CFL;
  float   lambda_padding;
  double  speed,  mu, rho;
  double  dt_Rho, dt_mu, Kwave_peak;
  double  B, z, one;

  double   xmin, xmax;
  double   h;
  double  *xv, *xs;
  double  *vel, *tao, *tao_initial;

  double dt;
  double tsimulations[NUM_REFLECTIONS];
  double pm[NUM_REFLECTIONS], em[NUM_REFLECTIONS];

  struct csr_matrix G;
  struct csr_matrix D;

  int curr_sim, step, i, j;

  char matdescra[6] = {'G', 'x', 'x', 'C', 'x', 'x'};
  const char transa = 'N';

  char *time_integration = "LF";

  /****** MPI job params ******/
  int    mpi_numtasks;           // Number of MPI tasks
  int    mpi_taskid;             // Task index
  int    p1, p2, p3, p4, p5, p6; // Parameter index
  int    tier1, tier2;
  int    counter;
  int    nparam  = 1000;   // Number of parameters
  int    totalparam;

  float  min_pm = 200.0;   // This task's minimum phase misfit
  float  min_em = 200.0;   // This task's minimum envelope misfit
  float *misfits;          // Array to store every resulting misfit

  float  param[2][nparam]; // Parameter space to explore
  float  minEnvelParam[2]; // Parameters for minimum envelope misfit
  float  minPhaseParam[2]; // Parameters for minimum phase misfit
  float  lowp;             // Parameter lower bound
  float  upp;              // Parameter upper bound
  float  dp;               // Parameter discretization

  /************************************************************************************/

  struct timeval  tv1, tv2;

  gettimeofday(&tv1, NULL);

  /************************************************************************************/

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD,&mpi_numtasks);
  MPI_Comm_rank(MPI_COMM_WORLD,&mpi_taskid);

  /* Initialize the parameter space */
  /* First alpha */
  lowp = -1.;
  upp  =  1.;
  dp = (upp - lowp) / (nparam-1);
  
  for(j = 0; j < nparam; j++)
    param[0][j] = j*dp + lowp;

  /* Second alpha */
  lowp = -1.;
  upp  =  1.;
  dp = (upp - lowp) / (nparam-1);

  for(j = 0; j < nparam; j++)
    param[1][j] = j*dp + lowp;

  /* Initialize the misfits for each combination of parameters */
  misfits = (float*) malloc(nparam*nparam*2*sizeof(float));

  for(p1 = 0;  p1 < nparam; p1++)
  for(p2 = 0;  p2 < nparam; p2++)
  {
    misfits[MISFIDX(p1,p2,0)] = 200.0;
    misfits[MISFIDX(p1,p2,1)] = 200.0;
  }

  /* Initialize propagation parameters */
  one = 1.0;

  CFL = 0.5;

  ppw            = 10;
  lambda_padding = 4.5;
  N = (1 + 2 * lambda_padding) * 2.5 * ppw;

  printf("************* Number of cells  : %d \n",N);

  speed = 3000.0;
  mu    = 30e9;
  rho   = mu / (speed*speed);

  for(i = 0; i < NUM_REFLECTIONS; i++)
    lambda_propagated[i] = reflections_times[i] * (1 + 2*lambda_padding);  // lambda_peak_propagated = [20 40 60 80 100 ...]

  printf("************* Lambda-Propagated: %d ",lambda_propagated[0]);
  for(i=1; i < NUM_REFLECTIONS; i++)
    printf(" %d ",lambda_propagated[i]);
  printf("\n");

  /************************************************************************************/

  /* Initialize the domain spatial properties */
  h    = 10.;                                  // grid spacing
  nx   = N + 1;
  nx1  = nx + 1; 
  xmin = 0; 
  xmax = N*h;                                  // domain size & spatial grid

  xv = (double *)malloc(nx  * sizeof(double));
  xs = (double *)malloc(nx1 * sizeof(double));

  xv[0] = xmin;
  xs[0] = xmin;

  for(i = 1; i < nx; i++)
  {
    xv[i] = i*h;                    // nx: velocity/displacement points   h = xv(2) - xv(1);  
    xs[i] = xv[i] - 0.5*h;          // ns: stress/displacement points     h = xv(2) - xv(1);
  }

  xs[nx] = xmax;

  /************************************************************************************/

  /* Initialize the domain time properties */
  dt = CFL*h/speed;                                // time step

  for(i=0; i < NUM_REFLECTIONS; i++)
  {
    tsimulations[i] = (xmax/speed)*reflections_times[i];   // Time instants of each intermediate simulations 
    nt[i] = (int) N * reflections_times[i]/CFL + 1;
  }

  if ( strcmp(time_integration,"LF") == 0 )
  {
    // Coefficients for O(2) Leapfrog  or explicit RK3 Integration   
     dt_Rho = dt / rho / h; 
     dt_mu = mu*dt / h; 
     z = sqrt(21);    
  }

  /************************************************************************************/
 
  /* Decomposition of the parameter space for 2 parameters */ 
  if(nparam < mpi_numtasks)
  {
    int stride = nparam * nparam / mpi_numtasks;   // Possible combinations per MPI task
    p1         = mpi_taskid / ( nparam / stride ); // Each value of the first parameter is assigned
    tier1      = p1 + 1;                           // to one MPI task
    p2         = (mpi_taskid % ( nparam / stride )) * (stride); // The remaining MPI tasks split the second
    tier2      = p2 + stride;                                   // parameter values among them.
  }
  else
  {
    int stride = nparam / mpi_numtasks;  // Divide the first parameter value space among 
    p1         = mpi_taskid * stride;    // the MPI tasks
    tier1      = p1 + stride; 
    p2         = 0;                      // Each MPI task will compute all values of the second
    tier2      = nparam;                 // parameter.
    printf("Rank=%d stride=%d p1=%d tier1=%d\n", mpi_taskid, stride, p1, tier1);
  }

  printf("************* Rank %d: P1[%.3f,%.3f]\n",mpi_taskid, param[0][p1], param[0][tier1-1]);
  printf("************* Rank %d: P2[%.3f,%.3f]\n",mpi_taskid, param[1][p2], param[1][tier2-1]);

  /* Monitoring */
  totalparam = (tier1-p1)*(tier2-p2); // Number of parameters to compute by the MPI task
  counter = 0;
  j = 0;

  for(p1 = p1; p1 < tier1;  p1++)
  for(p2 = p2; p2 < tier2;  p2++)
  {
    /* Init G and D operators */
    initGD44( &G, &D, N, N, param[0][p1], 0.0, -1./24., param[1][p2], 0.0, -1./24.); 
  
    /************************************************************************************/
  
    /* Initial pulse */
    vel         = (double *)calloc(nx,sizeof(double));
    tao         = (double *)calloc(N,sizeof(double));
    tao_initial = (double *)calloc(N,sizeof(double));
    
    Kwave_peak = 1/(2.5 * ppw * h);
  
    B = PI * PI * (Kwave_peak * Kwave_peak);
  
    for(i=1; i < nx; i++)
    { 
      double xr = (xs[i] - xmax/2.);
      double dU  = (1. - 2. * B * xr * xr ) * exp(- B * xr * xr);
      tao[i-1] = 0.001 * mu * dU;
    }
  
    memcpy(tao_initial, tao, N * sizeof(double));
  
    for(curr_sim=0; curr_sim < NUM_REFLECTIONS; curr_sim++)
    {
      /* Propagate the wave */
      for(step = 0; step < nt[curr_sim]; step++)
      {
        // vel =  vel  + dt_Rho*Gbar*tao;                                      
        mkl_dcsrmv(&transa, &nx , &N, &dt_Rho, matdescra, 
                   G.a , G.ja, G.ia , G.ia+1, tao , &one, vel );
  
        //tao = tao + dt_mu*D*vel;
        mkl_dcsrmv(&transa, &N, &nx, &dt_mu, matdescra, 
                   D.a , D.ja, D.ia, D.ia+1, vel, &one, tao );
      }

      /* Compare against initial pulse */
      pe_misfit(N,tao_initial,tao,&(pm[curr_sim]),&(em[curr_sim]));

      /* Save phase misfit */
      if(pm[curr_sim] < min_pm)
      {
        min_pm = pm[curr_sim];
        minPhaseParam[0] = param[0][p1];
        minPhaseParam[1] = param[1][p2];
      }

      /* Save envelope misfit */
      if(em[curr_sim] < min_em)
      {
        min_em = em[curr_sim];
        minEnvelParam[0] = param[0][p1];
        minEnvelParam[1] = param[1][p2];
      }

      misfits[MISFIDX(p1,p2,0)] = pm[curr_sim];
      misfits[MISFIDX(p1,p2,1)] = em[curr_sim];
    }

    counter++;

    /*  Root task reports progress on 5% increments */
    if( ((float)counter)/totalparam * 100 - j*5 > 5 && mpi_taskid == MPIROOT )
    {
      //printf("Counter=%d Total=%d j=%d\n", counter, totalparam, j);
      j++;
      printf("%d%% completed!\n", j*5);
    }
  }

  /************************************************************************************/
#if 0
  /* Write misfits to file */
  FILE *fp;
  char *mis_file;
  mis_file = (char *) malloc(128*sizeof(char));
  sprintf(mis_file,"./rmisfits_%d",mpi_taskid);
  fp = fopen(mis_file,"w+");

  fwrite(misfits, sizeof(float), nparam*nparam*nparam*nparam*nparam*nparam*2, fp);

  fclose(fp);
  free(mis_file);
  free(misfits);
#endif

  /* Report minimum misfits on current task */
  printf("RANK %d MINIMUM PHASE MISFIT=%f MINIMUM ENV MISFIT=%f\n", mpi_taskid, min_pm, min_em);
  if(min_pm >= 199.0 || min_em >= 199.0)
    printf("RANK %d unstable. No values found.\n",mpi_taskid);
  else
    printf("RANK %d vals= %.6f %.6f \n", mpi_taskid, minEnvelParam[0], minEnvelParam[1]);

  /************************************************************************************/

  MPI_Finalize();

  /************************************************************************************/

  gettimeofday(&tv2, NULL);

  printf ("Total time = %f seconds\n",
               (double) (tv2.tv_usec - tv1.tv_usec) / 1000000 +
                        (double) (tv2.tv_sec - tv1.tv_sec));

  return 0;
}


double vectors_dot_prod(const double *x, const double *y, int n)
{
    double res = 0.0;
    int i;
    for (i = 0; i < n; i++)
    {
        res += x[i] * y[i];
    }
    return res;
}

void matrix_vector_mult(const double *mat, const double *vec, double *result, int rows, int cols, double alpha)
{ // in matrix form: result += alpha * (mat * vec);
    int i;
    for (i = 0; i < rows; i++)
    {
        result[i] += alpha * vectors_dot_prod(&(mat[i*cols]), vec, cols);
    }
}
